/*************************************************************************
 *
 *  Test Examples for Libxac
 *
 *************************************************************************/
/*
 * Copyright (c) 2004-06    Massachusetts Institute of Technology
 * Copyright (c) 2004-06    Jim Sukha 
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */


/**
 * incSleepTest.c
 *
 *   Creates two processes that increment the same counter.
 *   Difference from incTest.c is that we have sleeps inserted.
 */



#include "libxac.h"
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

#define DEFAULT 100


int main(int argc, char* argv[]) {

  int answer;
  int count;
  int* test;
  int a0 = 0;
  int value;
  int status;
  int error;

  int numAborts = 0;
  int n = DEFAULT;
  int split;
  int finalAnswer;
  
  fprintf(stderr, "Running %s ... \n", argv[0]);


  if (argc >= 2) {
    n = atoi(argv[1]);
  }

  fprintf(stderr, "Running on input size %d\n", n);


  xInit(LIBXAC_DEFAULT_PATH, IS_DURABLE);
  test = (int*)xMmap(NULL, MAX_PAGES);

  
  error = 1;
  while (error != 0) {
    xbegin();
    test[a0] = 0;
    error = xend();
  }

  
  value = fork();

  
  if (value != 0) {
    split = n/2;
  }
  else {
    split = n - n/2;
  }
  
  for (count = 0; count < split; count++) {

    error = 1;
    while (error != 0) {
      int temp = 0;
      xbegin();
      temp = test[a0];
      usleep(10);
      temp++;
      
      test[a0] = temp;
      error = xend();

      if (error != 0) {
	//	fprintf(stderr, "Process %d has aborted %d times. PID is %d\n", value, numAborts, getpid());
	numAborts++;
      }
    }
  }

  error = 1;
  while (error != 0) {
    xbegin();
    fprintf(stderr, "Interior final value for process %d: %d\n", value, test[a0]);
    answer = test[a0];
    error = xend();
  }
    

  fprintf(stderr, "Final value for process %d: %d\n", getpid(), answer);
  fprintf(stderr, "Process %d aborted %d times. \n", getpid(), numAborts);


  if (value != 0) {

    while (wait(&status) == -1) {
    }

    error = 1;
    while (error != 0) {
      xbegin();
      fprintf(stderr, "\nTotal final value: %d\n", test[a0]);
      finalAnswer = test[a0];
      error = xend();
    }
  
    //    printf("Errno is %d\n", errno);
    //    printf("ECHILD is %d\n", ECHILD);
    //    printf("EINTR is %d\n", EINTR);

    xMunmap(NULL);
    xShutdown();

    if (finalAnswer != n) {
      fprintf(stderr, "Incorrect answer for incTest.exe. Answer is %d instead of %d\n",
	      finalAnswer,
	      n);
      assert(FALSE);
      return 1;
    }
    else {

      fprintf(stderr, "%s returned CORRECT value of %d.\n", argv[0], finalAnswer);
      return 0;
    }
  }
  else {
    xMunmap(NULL);
  }

  return 0;
}



