/* runtime/cilk-sysdep.h.  Generated by configure.  */
/* -*- C -*-
 */
/*
 * Copyright (c) 2000 Massachusetts Institute of Technology
 * Copyright (c) 2000 Matteo Frigo
 *
 *  This library is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation; either version 2.1 of the License, or (at
 *  your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
 *  USA.
 *
 */

/**
 * I borrowed this from Cilk to get locks that work.
 * Eventually, this needs to be changed to run on other platforms...
 * Hard-coded for x86 at the moment...
 *  
 *   Jim
 */
static const char *ident_cilk_sysdep_h __attribute__((__unused__))
     = "$HeadURL: https://bradley.csail.mit.edu/svn/repos/cilk/current.atomic/runtime/cilk-sysdep.h.in $ $LastChangedBy: sukhaj $ $Rev: 771 $ $Date: 2003-10-27 09:54:49 -0500 (Mon, 27 Oct 2003) $";

/* Define if you have the gethrtime function.  */
/* #undef HAVE_GETHRTIME */

/* Define if you have the gethrvtime function.  */
/* #undef HAVE_GETHRVTIME */

/* Define if you have the gettimeofday function.  */
#define HAVE_GETTIMEOFDAY 1

/* Define if you have the <sys/time.h> header file.  */
#define HAVE_SYS_TIME_H 1

/* Define if you have the <unistd.h> header file.  */
#define HAVE_UNISTD_H 1

/* Define if you have the memalign function.  */
#define HAVE_MEMALIGN 1

/***********************************************************\
 * OS-specific macros
\***********************************************************/

#if defined(__osf__) && defined(__alpha__)
/* c_asm.h declares a variable called `asm' which crashes gcc */
#  ifndef __C_ASM_H
#    define __C_ASM_H
#  endif
#endif

/***********************************************************\
 * Various types of memory barriers and atomic operations
\***********************************************************/

/*------------------------
       POWERPC
 ------------------------*/
#if defined(__powerpc__) || defined(__ppc__)
#  define CILK_CACHE_LINE 64
#  define CILK_MB()  __asm__ __volatile__ ("sync" : : : "memory")
#  define CILK_RMB()  __asm__ __volatile__ ("sync" : : : "memory")
#  define CILK_WMB()  __asm__ __volatile__ ("eieio" : : : "memory")

   /* atomic swap operation */
   static inline int Cilk_xchg(volatile int *ptr, int x)
   {
	int result;
	__asm__ __volatile__ (
	   "0: lwarx %0,0,%1\n stwcx. %2,0,%1\n bne- 0b" :
	    "=&r"(result) :
	   "r"(ptr), "r"(x) :
	    "cr0");
	return result;
   }
#endif

/*------------------------
       IA64
 ------------------------*/
#ifdef __ia64__
#  define CILK_CACHE_LINE 128 /* just a guess */
#  define CILK_MB()  __asm__ __volatile__ ("mf" : : : "memory")
#  define CILK_RMB()  CILK_MB()
#  define CILK_WMB()  CILK_MB()

   /* atomic swap operation */
   static inline int Cilk_xchg(volatile int *ptr, int x)
   {
	int result;
	__asm__ __volatile ("xchg4 %0=%1,%2" : "=r" (result)
			    : "m" (*(int *) ptr), "r" (x) : "memory");
	return result;
   }
#endif

/*------------------------
         I386
 ------------------------*/
#ifdef __i386__
#  define CILK_CACHE_LINE 32

   static inline void CILK_MB(void)
   {
	/*
	 * We use an xchg instruction to serialize memory accesses, as can
	 * be done according to the Intel Architecture Software Developer's
	 * Manual, Volume 3: System Programming Guide
	 * (http://www.intel.com/design/pro/manuals/243192.htm), page 7-6,
	 * "For the P6 family processors, locked operations serialize all
	 * outstanding load and store operations (that is, wait for them to
	 * complete)."  The xchg instruction is a locked operation by
	 * default.  Note that the recommended memory barrier is the cpuid
	 * instruction, which is really slow (~70 cycles).  In contrast,
	 * xchg is only about 23 cycles (plus a few per write buffer
	 * entry?).  Still slow, but the best I can find.  -KHR
	 *
	 * Bradley also timed "mfence", and on a Pentium IV xchgl is still quite a bit faster
	 *   mfence appears to take about 125 ns on a 2.5GHZ P4
	 *   xchgl  apears  to take about  90 ns on a 2.5GHZ P4
	 * However on an opteron, the performance of mfence and xchgl are both *MUCH MUCH BETTER*.
	 *   mfence takes 8ns on a 1.5GHZ AMD64 (maybe this is an 801)
	 *   sfence takes 5ns
	 *   lfence takes 3ns
	 *   xchgl  takes 14ns
	 * see mfence-benchmark.c
	 */
	int x=0, y;
	asm volatile ("xchgl %0,%1" :"=r" (x) :"m" (y), "0" (x) :"memory");
   }

#  define CILK_RMB() CILK_MB()
#  define CILK_WMB() __asm__ __volatile__ ("": : :"memory")

   /* atomic swap operation */
   static inline int Cilk_xchg(volatile int *ptr, int x)
   {
	__asm__("xchgl %0,%1" :"=r" (x) :"m" (*(ptr)), "0" (x) :"memory");
	return x;
   }
#endif /* __i386__ */

/*------------------------
         amd_64
 ------------------------*/
#ifdef __x86_64
#  define CILK_CACHE_LINE 64

   static inline void CILK_MB(void) 
   {
     asm volatile ("mfence": : :"memory");
   }
#  define CILK_RMB() CILK_MB()
#  define CILK_WMB() __asm__ __volatile__ ("sfence": : :"memory")

   /* atomic swap operation */
   static inline int Cilk_xchg(volatile int *ptr, int x)
   {
	__asm__("xchgl %0,%1" :"=r" (x) :"m" (*(ptr)), "0" (x) :"memory");
	return x;
   }
#endif /* __x86_64 */

/*------------------------
         SPARC
 ------------------------*/
#ifdef __sparc__

#define CILK_CACHE_LINE 64
   /* atomic swap operation */
   static inline int Cilk_xchg(volatile int *ptr, register int xchg)
   {
	asm volatile ("swap [%1],%0"
		      :"=r" (xchg)	/* registers written */
		      :"r"(ptr), "0"(xchg)		/* registers read */
		      : "memory"
	);
	return xchg;
   }


#  ifdef __sparc_v9__
#    define CILK_MB()  \
       __asm__ __volatile__("membar #StoreLoad | #StoreStore | #LoadLoad | #LoadStore" ::: "memory")
#    define CILK_RMB() __asm__ __volatile__ ("membar #StoreLoad" ::: "memory")
#    define CILK_WMB() __asm__ __volatile__ ("": : :"memory")
#  else
#    define CILK_MB()  __asm__ __volatile__("stbar" ::: "memory")
     static inline void CILK_RMB(void) { int dummy; Cilk_xchg(&dummy, 0); }
#    define CILK_WMB() __asm__ __volatile__ ("": : :"memory")
#endif

#endif /* __sparc__ */

/*------------------------
         MIPS
 ------------------------*/
#ifdef __mips__
#  define CILK_CACHE_LINE 128
#  define CILK_MB() __asm__ __volatile__ ("sync":::"memory")
#  define CILK_RMB() CILK_MB()
#  define CILK_WMB() CILK_MB()

   /* atomic swap operation */
   static inline int Cilk_xchg(volatile int *m, int val)
   {
	unsigned long dummy;
	__asm__ __volatile__(
	     ".set noreorder\n"
	     ".set noat\n"
	     "ll %0,(%1)\n"
	     "1: move $1,%2\n"
	     "sc $1,(%1)\n"
	     "beqzl $1,1b\n"
	     "ll %0,(%1)\n"
	     ".set at\n"
	     ".set reorder"
	     : "=r" (val), "=r" (m), "=r" (dummy)
	     : "1" (m), "2" (val)
	     : "memory");
	return val;
   }
#endif /* __mips__ */

/*------------------------
         ALPHA
 ------------------------*/
#ifdef __alpha__
#  define CILK_CACHE_LINE 128
#  define CILK_MB() __asm__ __volatile__ ("mb":::"memory")
#  define CILK_RMB() __asm__ __volatile__ ("mb":::"memory")
#  define CILK_WMB() __asm__ __volatile__ ("wmb":::"memory")

   /* atomic swap operation */
   static inline int Cilk_xchg(volatile int *m, int val)
   {
	register int dummy;

	__asm__ __volatile__(
	     "1:    ldl_l %0,%2\n"
	     "      bis $31,%3,%1\n"
	     "	    stl_c %1,%2\n"
	     "	    beq %1,1b\n"
	     "	    mb\n"
	     : "=&r" (val), "=&r" (dummy), "=m" (*m)
	     : "rI" (val), "m" (*m));

	return val;
   }
#endif /* __alpha__ */

#ifndef CILK_CACHE_LINE
#  error "Unsupported CPU"
#endif

/*
 * padding, to be used in structures to avoid false sharing
 */
#define CILK_CACHE_LINE_PAD  char __dummy[CILK_CACHE_LINE]
#define CILK_CACHE_LINE_ALIGNED(decl) \
    decl  __attribute__((aligned(CILK_CACHE_LINE)))

/***********************************************************\
 * Timing functions
\***********************************************************/
#if defined(HAVE_GETHRTIME)

/* we use the nanosecond virtual timer */
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

typedef hrtime_t Cilk_time;

#ifdef HAVE_GETHRVTIME
static inline Cilk_time Cilk_get_time(void)
{
     return gethrvtime();
}
#else
static inline Cilk_time Cilk_get_time(void)
{
     return gethrtime();
}
#endif

static inline double Cilk_time_to_sec(Cilk_time t)
{
     return (double) t / 1.0e9;
}

static inline Cilk_time Cilk_get_wall_time(void)
{
     return gethrtime();
}

static inline double Cilk_wall_time_to_sec(Cilk_time t)
{
     return (double) t / 1.0e9;
}

#elif defined(HAVE_GETTIMEOFDAY)
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

typedef long long Cilk_time;

static inline Cilk_time Cilk_get_time(void)
{
     struct timeval tv;
     gettimeofday(&tv, 0);

     return ((Cilk_time)tv.tv_sec * 1000000LL + (Cilk_time)tv.tv_usec);
}

static inline double Cilk_time_to_sec(Cilk_time t)
{
     return (double) t * 1.0E-6;
}

static inline Cilk_time Cilk_get_wall_time(void)
{
     return Cilk_get_time();
}

static inline double Cilk_wall_time_to_sec(Cilk_time t)
{
     return Cilk_time_to_sec(t);
}

#endif
