/*************************************************************************
 *
 *  B-tree code -- utilities
 *
 *************************************************************************/
/*
 * Copyright (c) 2004-06    Massachusetts Institute of Technology
 * Copyright (c) 2004-06    Bradley C. Kuszmaul
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

/* Useful utilities that show up over and over... */

#include <errno.h>
#include <stdlib.h>
#include <stdio.h>

#include "utillib.h"

/* return 0 if the string can be parsed into an unsigned long.
 * Things that are considered bad:
 *   an empty input string
 *   overflow
 *   garbage at the beginning or end of the string
 */
int strtoul_carefully (/* in */
		       const char *string,
		       /* out */
		       unsigned long *result) {
  char *end;
  unsigned long r;
  errno=0;
  r = strtoul(string, &end, 10);
  if (errno!=0) return errno;  /* overflow */
  if (end==string) return -1;  /* garbage at the beginning, or empty string */
  if (*end) return -1;         /* garbage at end */
  *result=r;
  return 0;
}

unsigned long strtoul_or_error (const char *string, const char *error_message) {
  unsigned long result;
  if (strtoul_carefully(string, &result)) {
    fprintf(stderr, "parsing error:  Trying to parse \"%s\" as an unsigned long: %s\n", string, error_message);
    exit(1);
  }
  return result;
}
